package meerkat;

import com.google.protobuf.ByteString;
import com.google.protobuf.InvalidProtocolBufferException;
import meerkat.ProtobufsMessages.BasicMessage;
import meerkat.ProtobufsMessages.Tag;
import meerkat.bulletinboard.SimpleBulletinBoardClient;
import meerkat.comm.CommunicationException;
import meerkat.crypto.Encryption;
import meerkat.protobuf.BulletinBoardAPI;
import meerkat.protobuf.Crypto;
import util.AccurateTimestamp;
import util.CollectionMessagesUtils;
import util.Tags;

import java.io.IOException;
import java.text.ParseException;
import java.util.*;

/**
 * Created by Vladimir Eliezer Tokarev on 1/8/2016.
 * Gives the ability to manage voters information
 */
public class SimpleRegistry {

    protected Encryption signatory;

    protected SimpleBulletinBoardClient communicator;

    public SimpleRegistry(Encryption signatory, SimpleBulletinBoardClient communicator) {
        this.signatory = signatory;
        this.communicator = communicator;
    }

    private Tag.Builder[] GetTagsArrayInLength(int length) {
        Tag.Builder[] tags = new Tag.Builder[length];

        for (int i = 0; i < length; i++) {
            tags[i] = Tag.newBuilder();
        }
        return tags;
    }

    /**
     * Creates BulletinBoardMessage with signed basicMessage and UnsignedBulletinBoardMessage that contains the basic message
     *
     * @param basicMessage BasicMessage
     * @return BulletinBoardAPI.BulletinBoardMessage
     */
    private BulletinBoardAPI.BulletinBoardMessage CreateBulletinBoardMessage(BasicMessage basicMessage) throws IOException {
        BulletinBoardAPI.BulletinBoardMessage.Builder bulletinBoardMessage =
                BulletinBoardAPI.BulletinBoardMessage.newBuilder();

        Crypto.RerandomizableEncryptedMessage encryptedMessage =
                signatory.encrypt(basicMessage, signatory.generateRandomness(new Random()));

        Crypto.Signature.Builder messageSignature = Crypto.Signature.newBuilder();
        messageSignature.mergeFrom(encryptedMessage);

        BulletinBoardAPI.UnsignedBulletinBoardMessage.Builder unsignedBulletinBoardMessage =
                BulletinBoardAPI.UnsignedBulletinBoardMessage.newBuilder();
        unsignedBulletinBoardMessage.setData(basicMessage.toByteString());

        bulletinBoardMessage.addSig(messageSignature);

        bulletinBoardMessage.setMsg(unsignedBulletinBoardMessage);

        return bulletinBoardMessage.build();
    }

    /**
     * Gets messages that have the wanted id tag and have or the ADD_TO_GROUP_TAG or REMOVE_FROM_GROUP_TAG
     *
     * @param tags the tags based on which the messages will be filtered
     * @return List<BulletinBoardAPI.BulletinBoardMessage>
     */
    private List<BulletinBoardAPI.BulletinBoardMessage> GetRelevantMessages(List<Tag> tags) {
        BulletinBoardAPI.MessageFilterList.Builder filters =
                BulletinBoardAPI.MessageFilterList.newBuilder();

        for (Tag tag : tags) {
            BulletinBoardAPI.MessageFilter.Builder idFilter =
                    BulletinBoardAPI.MessageFilter.newBuilder().setTag(tag.getContent());

            filters.addFilter(idFilter);
        }
        return communicator.readMessages(filters.build());
    }

    /**
     * Gets Relevant bulletinBoard messages from communicator than converts them to VoterRegistryMessages
     *
     * @param tags list of tags that will be used as filters
     * @return List<VoterRegistryMessage>
     */
    private List<VoterRegistryMessage> GetRelevantVoterRegistryMessages(List<Tag> tags) throws InvalidProtocolBufferException {
        List<BulletinBoardAPI.BulletinBoardMessage> relevantMessages = GetRelevantMessages(tags);
        List<BasicMessage> messages = CollectionMessagesUtils.GetBasicMessagesFromBulletinBoardMessages(relevantMessages);
        List<VoterRegistryMessage> voterRegistryMessages = CollectionMessagesUtils.ConvertToVoterRegistryMessages(messages);

        return voterRegistryMessages;
    }

    /**
     * Adds new voter to the bulletin-board
     *
     * @param voterID
     * @param personalData for example residence location
     * @return void
     * @throws throws CommunicationException
     */
    public void SetVoted(String voterID, String personalData) throws CommunicationException, IOException {
        Tag.Builder[] tags = GetTagsArrayInLength(3);
        tags[0].setContent(Tags.ID_TAG + " " + voterID);
        tags[1].setContent(Tags.VOTER_ENTRY_TAG.toString());
        tags[2].setContent(Tags.ACTION_TIMESTAMP_TAG + " " + AccurateTimestamp.GetCurrentTimestampString());

        BasicMessage.Builder basicMessage = BasicMessage.newBuilder().addTag(tags[0]).addTag(tags[1]).addTag(tags[1]);
        basicMessage.setData(ByteString.copyFrom(personalData.getBytes()));

        communicator.postMessage(CreateBulletinBoardMessage(basicMessage.build()));
    }

    /**
     * Adding given voter to given group
     *
     * @param voterID
     * @param groupID
     * @return true if the adding action succeeded else return false
     * @throws CommunicationException
     */
    public void AddToGroup(String voterID, String groupID) throws CommunicationException, IOException {
        Tag.Builder[] tags = GetTagsArrayInLength(4);
        tags[0].setContent(Tags.ID_TAG + " " + voterID);
        tags[1].setContent(Tags.GROUP_ID_TAG + " " + groupID);
        tags[2].setContent(Tags.GROUP_ACTION_TAG + " " + Tags.ADD_TO_GROUP_TAG);
        tags[3].setContent(Tags.ACTION_TIMESTAMP_TAG + " " + AccurateTimestamp.GetCurrentTimestampString());

        BasicMessage.Builder basicMessage =
                BasicMessage.newBuilder().addTag(tags[0]).addTag(tags[1]).addTag(tags[2]).addTag(tags[3]);

        communicator.postMessage(CreateBulletinBoardMessage(basicMessage.build()));
    }

    /**
     * Removes given voter from given group
     *
     * @param voterID
     * @param groupID
     * @return true if the removing action succeeded else return false
     * @throws CommunicationException
     */
    public void RemoveFromGroup(String voterID, String groupID) throws CommunicationException, IOException {
        Tag.Builder[] tags = GetTagsArrayInLength(4);
        tags[0].setContent(Tags.ID_TAG + " " + voterID);
        tags[1].setContent(Tags.GROUP_ID_TAG + " " + groupID);
        tags[2].setContent(Tags.GROUP_ACTION_TAG + " " + Tags.REMOVE_FROM_GROUP_TAG);
        tags[3].setContent(Tags.ACTION_TIMESTAMP_TAG + " " + AccurateTimestamp.GetCurrentTimestampString());

        BasicMessage.Builder basicMessage =
                BasicMessage.newBuilder().addTag(tags[0]).addTag(tags[1]).addTag(tags[2]).addTag(tags[3]);

        communicator.postMessage(CreateBulletinBoardMessage(basicMessage.build()));
    }

    /**
     * Sets that the voter have voted
     *
     * @param id id tag string
     * @return true if the set voted succeded else false
     * @throws CommunicationException
     */
    public void SetVoted(String id) throws CommunicationException, IOException {
        Tag.Builder[] tags = GetTagsArrayInLength(3);
        tags[0].setContent(Tags.ID_TAG + " " + id);
        tags[1].setContent(Tags.VOTE_ACTION_TAG.toString());
        tags[2].setContent(Tags.ACTION_TIMESTAMP_TAG + " " + AccurateTimestamp.GetCurrentTimestampString());

        BasicMessage.Builder basicMessage = BasicMessage.newBuilder().addTag(tags[0]).addTag(tags[1]).addTag(tags[2]);

        communicator.postMessage(CreateBulletinBoardMessage(basicMessage.build()));
    }


    /**
     * Requests all the groups that the given id voter is in
     *
     * @param id id tag string
     * @return list of groups ids (or names), if the method fails its empty
     * @throws CommunicationException, InvalidProtocolBufferException
     */
    public List<String> GetGroups(String id) throws CommunicationException, InvalidProtocolBufferException {
        List<Tag> GroupsActionsTags = new ArrayList<Tag>();
        GroupsActionsTags.add(Tag.newBuilder().setContent(Tags.ID_TAG + " " + id).build());
        GroupsActionsTags.add(Tag.newBuilder().setContent(Tags.GROUP_ACTION_TAG + " " + Tags.REMOVE_FROM_GROUP_TAG).build());
        GroupsActionsTags.add(Tag.newBuilder().setContent(Tags.GROUP_ACTION_TAG + " " + Tags.ADD_TO_GROUP_TAG).build());

        List<VoterRegistryMessage> voterRegistryMessages = GetRelevantVoterRegistryMessages(GroupsActionsTags);

        try {
            Map<String, VoterRegistryMessage> groupIdToMessage = CollectionMessagesUtils.GetLatestGroupsActions(voterRegistryMessages);
            return CollectionMessagesUtils.GetListOfGroupIds(groupIdToMessage);
        } catch (ParseException e) {
            return null;
        }
    }

    /**
     * Retrieves list of strings that represents voter
     *
     * @param id id tag string
     * @return list of strings (empty list if the lookup failed)
     * @throws CommunicationException
     */
    public List<String> GetPersonIDDetails(String id) throws CommunicationException, InvalidProtocolBufferException, ParseException {
        List<Tag> GroupsActionsTags = new ArrayList<Tag>();
        GroupsActionsTags.add(Tag.newBuilder().setContent(Tags.ID_TAG + " " + id).build());
        GroupsActionsTags.add(Tag.newBuilder().setContent(Tags.VOTER_ENTRY_TAG.toString()).build());

        List<VoterRegistryMessage> voterRegistryMessages = GetRelevantVoterRegistryMessages(GroupsActionsTags);
        VoterRegistryMessage LatestMessage = voterRegistryMessages.get(0);

        for (VoterRegistryMessage message : voterRegistryMessages) {
            if (LatestMessage.GetBasicMessageActionTimestamp().before(message.GetBasicMessageActionTimestamp()) {
                LatestMessage = message;
            }
        }

        return Arrays.asList(LatestMessage.GetWantedTagFromBasicMessage(Tags.ID_TAG), LatestMessage.base.getData().toString());
    }
}
